
import java.io.*;
import java.util.Date;
import java.net.*;

public class Gameboy extends GBZ80 {
  
  final static byte NORAM  =   0x00;      /* Byte to be returned from      */
                            /* non-existing pages and ports  */
  final static int MAXCHEAT = 256;       /* Maximal number of GG cheats   */

  final static byte VBL_IFLAG = 0x01;
  final static byte LCD_IFLAG = 0x02;
  final static byte TIM_IFLAG = 0x04;
  final static byte SIO_IFLAG = 0x08;
  final static byte EXT_IFLAG = 0x10;
 
  final static char JOYPAD  = (char) 0xFF00; /* Joystick: 1.1.P15.P14.P13.P12.P11.P10      */
  final static char SIODATA = (char) 0xFF01; /* Serial IO data buffer                      */
  final static char SIOCONT = (char) 0xFF02; /* Serial IO control register                 */
  final static char DIVREG  = (char) 0xFF04; /* Divider register (???)                     */
  final static char TIMECNT = (char) 0xFF05; /* Timer counter. Gen. int. when it overflows */
  final static char TIMEMOD = (char) 0xFF06; /* New value of TimeCount after it overflows  */
  final static char TIMEFRQ = (char) 0xFF07; /* Timer frequency and start/stop switch      */
  final static char IFLAGS  = (char) 0xFF0F; /* Interrupt flags: 0.0.0.JST.SIO.TIM.LCD.VBL */
  final static char ISWITCH = (char) 0xFFFF; /* Switches to enable/disable interrupts      */
  final static char LCDCONT = (char) 0xFF40; /* LCD control register                       */
  final static char LCDSTAT = (char) 0xFF41; /* LCD status register                        */
  final static char SCROLLY = (char) 0xFF42; /* Starting Y position of the background      */
  final static char SCROLLX = (char) 0xFF43; /* Starting X position of the background      */
  final static char CURLINE = (char) 0xFF44; /* Current screen line being scanned          */
  final static char CMPLINE = (char) 0xFF45; /* Gen. int. when scan reaches this line      */
  final static char BGRDPAL = (char) 0xFF47; /* Background palette                         */
  final static char SPR0PAL = (char) 0xFF48; /* Sprite palette #0                          */
  final static char SPR1PAL = (char) 0xFF49; /* Sprite palette #1                          */
  final static char WNDPOSY = (char) 0xFF4A; /* Window Y position                          */
  final static char WNDPOSX = (char) 0xFF4B; /* Window X position                          */

  final static int visWidth = 160;
  final static int visStartX = 8;
  final static int WIDTH  = 176;  /* Width (>=176, must be divisible by 8) */
  final static int HEIGHT = 144;  /* Height (>=144)                        */
  byte[] XBuf= new byte[HEIGHT*WIDTH];
  byte[] ZBuf= new byte[2*WIDTH*HEIGHT/8];
  byte[] XPal = new byte[12];

  byte Verbose    = 1;     /* Verboseness level                            */

  byte[][] RAM = new byte[8][0x2000];
  byte[][] Page = new byte[8][];      /* RAM and pointers to Z80 address space 8x8kB) */

  // use tstatesPerInterrupt instead int  IPeriod; /* Number of cycles between internal interrupts */
  boolean IntSync;  /* Generate internal interrupts if IntSync==1   */
  boolean IFlag;    /* If IFlag==1, generate interrupt and set to 0 */
  
  int  VPeriod    = 6000;  /* Number of Z80 cycles between VBlanks         */
  int UPeriod    = 2;     /* Number of VBlanks per screen update          */
  boolean LineDelay  = false;     /* When 1, CMPLINE interrupts are delayed       */
  boolean CheckCRC   = true;     /* When 1, VGB checks cartridge CRC on loading  */
  boolean AutoA      = false;     /* When 1, autofire emulation for button A      */
  boolean AutoB      = false;     /* When 1, autofire emulation for button B      */

  String SaveName  = null;  /* .sav file name                               */

  String SndName   = null;  /* Soundtrack log file                          */
  FileOutputStream SndStream = null;

  // struct                   /* Cheat list to be filled before StartGB()     */
  // {                        /* is called                                    */
  //   byte Value,Orig;
  //   word Address;
  // } Cheats[MAXCHEAT];
  int CheatCount  = 0;     /* Number of cheats in the list                 */

  byte SIOBuf;             /* Next byte to output via the serial line      */
  byte SIOFlag;            /* Bit0 -> SIO interrupt should be generated    */
                           /* Bit1 -> Data should be sent out              */

  byte JoyState= (byte) 0xff;      /* (1 = not pressed) Joystick state: START.SELECT.B.A.D.U.L.R     */

  byte IMask;              /* A mask to reset an event bit in IFLAGS       */

  boolean MBCType;            /* MBC type: 1 for MBC2, 0 for MBC1             */
  // The C version had 16K ROM Banks. But we need to have them be 8K banks so
  // that we can easily put them in the Pages[] array.
  byte ROMMap[][] = new byte[512][];       /* Addresses of ROM banks                       */
  byte ROMBank;            /* Number of ROM bank currently used            */
  byte ROMMask;            /* Mask for the ROM bank number                 */
  int  ROMBanks;           /* Total number of ROM banks                    */
  byte RAMMap[][] = new byte[256][];       /* Addresses of RAM banks                       */ 
  byte RAMBank;            /* Number of RAM bank currently used            */
  byte RAMMask;            /* Mask for the RAM bank number                 */
  int  RAMBanks;           /* Total number of RAM banks                    */

  int TCount,TStep;  /* Timer counter and increment              */

  byte[] BPal = new byte[4];            /* Background palette                           */
  byte[][] SPal = new byte[2][4]; /* Sprite palettes                              */

  // Offset into RAM of these objects
  char ChrGen;    /* Character generator                          */
  char BgdTab;
  char WndTab;    /* Background and window character tables       */

  byte SprFlag;            /* <>0: sprites were enabled during the frame   */
  
  GameboyObserver fMyObserver; // So we don't know about our clients.
  
  public Gameboy(){
  }

  String getCompanyForCode(int code){
    switch(code & 0xffff) {
    case  0x3301: return "Nintendo"  ;case  0x7901: return "Accolade"  ;case  0xA400: return "Konami"    ;
    case  0x6701: return "Ocean"     ;case  0x5601: return "LJN"       ;case  0x9900: return "ARC?"      ;
    case  0x0101: return "Nintendo"  ;case  0x0801: return "Capcom"    ;case  0x0100: return "Nintendo"  ;
    case  0xBB01: return "SunSoft"   ;case  0xA401: return "Konami"    ;case  0xAF01: return "Namcot?"   ;
    case  0x4901: return "Irem"      ;case  0x9C01: return "Imagineer" ;case  0xA600: return "Kawada?"   ;
    case  0xB101: return "Nexoft"    ;case  0x5101: return "Acclaim"   ;case  0x6001: return "Titus"     ;
    case  0xB601: return "HAL"       ;case  0x3300: return "Nintendo"  ;case  0x0B00: return "Coconuts?" ;
    case  0x5401: return "Gametek"   ;case  0x7F01: return "Kemco?"    ;case  0xC001: return "Taito"     ;
    case  0xEB01: return "Atlus"     ;case  0xE800: return "Asmik?"    ;case  0xDA00: return "Tomy?"     ;
    case  0xB100: return "ASCII?"    ;case  0xEB00: return "Atlus"     ;case  0xC000: return "Taito"     ;
    case  0x9C00: return "Imagineer" ;case  0xC201: return "Kemco?"    ;case  0xD101: return "Sofel?"    ;
    case  0x6101: return "Virgin"    ;case  0xBB00: return "SunSoft"   ;case  0xCE01: return "FCI?"      ;
    case  0xB400: return "Enix?"     ;case  0xBD01: return "Imagesoft" ;case  0x0A01: return "Jaleco?"   ;
    case  0xDF00: return "Altron?"   ;case  0xA700: return "Takara?"   ;case  0xEE00: return "IGS?"      ;
    case  0x8300: return "Lozc?"     ;case  0x5001: return "Absolute?" ;case  0xDD00: return "NCS?"      ;
    case  0xE500: return "Epoch?"    ;case  0xCB00: return "VAP?"      ;case  0x8C00: return "Vic Tokai" ;
    case  0xC200: return "Kemco?"    ;case  0xBF00: return "Sammy?"    ;
    case  0x1800: return "Hudson Soft"    ;case  0xCA01: return "Palcom/Ultra" ;
    case  0xCA00: return "Palcom/Ultra"   ;case  0xC500: return "Data East?" ;
    case  0xA900: return "Technos Japan?" ;case  0xD900: return "Banpresto?" ;
    case  0x7201: return "Broderbund?"    ;case  0x7A01: return "Triffix Entertainment?" ;
    case  0xE100: return "Towachiki?"     ;case  0x9300: return "Tsuburava?" ;
    case  0xC600: return "Tonkin House?"  ;case  0xCE00: return "Pony Canyon" ;
    case  0x7001: return "Infogrames?"    ;case  0x8B01: return "Bullet-Proof Software?" ;
    case  0x5501: return "Park Place?"    ;case  0xEA00: return "King Records?" ;
    case  0x5D01: return "Tradewest?"     ;case  0x6F01: return "ElectroBrain?" ;
    case  0xAA01: return "Broderbund?"    ;case  0xC301: return "SquareSoft" ;
    case  0x5201: return "Activision?"    ;case  0x5A01: return "Bitmap Brothers/Mindscape" ;
    case  0x5301: return "American Sammy" ;case  0x4701: return "Spectrum Holobyte" ;
    case  0x1801: return "Hudson Soft";
    default:
      return null;
    }
  };

  final int M_RDMEM(int A)
  { return (0xff & (Page[A>>13][A&0x1FFF])); }

  final void M_WRMEM(int A, int V)
  {
    if(A>0xFEFF) DoWrite(A,V);
    else if(A>0x7FFF) Page[A>>13][A&0x1FFF]=(byte) V;
         else

    switch(A&0xE000)
    {
      case 0xC000:
      case 0x8000: RAM[A>>13][A&0x1FFF]=(byte) V;return;
      case 0xA000: Page[5][A&0x1FFF]=(byte) V;return;
      case 0xE000: DoWrite(A,V);return;
      case 0x2000:
        if(MBCType &&((A&0xFF00)!=0x2100)) return;
        V&=ROMMask;
        if(V == 0) V++;
        if((ROMMask != 0)&&(V!=ROMBank))
        {
          ROMBank=(byte) V;
          Page[2]=ROMMap[V*2] != null ? ROMMap[V*2]:RAM[0x4000>>13];
          // The code originally was Page[3]=Page[2]+0x2000
          Page[3]=ROMMap[V*2+1] != null ? ROMMap[V*2+1]:RAM[0x6000>>13];
          if((Verbose&0x08) != 0) System.out.println("ROM: Bank " + hex2(V) + " selected");
        }
        return;
      case 0x4000:
        V&=RAMMask;
        if(RAMMask != 0 &&!MBCType &&(RAMBank!=V))
        {
          RAMBank=(byte) V;
          Page[5]= (RAMMap[V] != null)? RAMMap[V]:RAM[0xA000>>13];
          if((Verbose&0x08) != 0) System.out.println("RAM: Bank " + hex2(V) + " selected");
        }
        return;
      case 0x0000:
      case 0x6000:
        if((Verbose&0x02) != 0) System.out.println("Wrote " + hex2(V) + " to ROM at " + hex4(A));
        // Trace(2);
        return;
    }
  }

  private final int TPFreqs[] = { 4096,262144,65536,16384 };
  void DoWrite( int A, int V)
  {
    // System.err.println(Debug.hex4(A) +": " + Debug.hex2(V));
    switch(A)
    {
      case 0xFF00: primitiveStore(JOYPAD, 0xCF|V);
                   if((V&0x20) == 0){
                     int value = (((0xff & JoyState)>>4)|0xF0);
                     RAM[JOYPAD>>13][JOYPAD&0x1fff] &= (byte) value;
                   }
                   if((V&0x10) == 0){
                     int value = ((0xff & JoyState)|0xF0);
                     RAM[JOYPAD>>13][JOYPAD&0x1fff] &= (byte) value; 
                   }
                   return;
      case 0xFF01: SIOBuf= (byte) V;SIOFlag|=0x02;return;
      case 0xFF02: if((V&0x80) != 0)
                   {
                     if(((SIOFlag&0x02) != 0 ? SIOSend():SIOReceive()) != 0)
                       SIOFlag|= (byte) 0x01;
                     SIOFlag=(byte) (SIOFlag&0xFD);
                   }
                   V|=0x7E; 
                   break;
      case 0xFF07: TStep=(TPFreqs[V&0x03]<<16)/(154*60);
                   V|=0xF8;break;
      case 0xFF0F: V&=0x1F;break;
      case 0xFFFF: V&=0x1F;break;
      case 0xFF46: {
                      // Copy 0xA0 bytes from anywhere in RAM (on a page boundary)
                      // to the OAM (the Sprite Attribute Table).
                      
                      final int kDest = 0xFE00;
                      byte[] dest = RAM[kDest>>13];
                      final int p = kDest & 0x1fff;
                      
                      int kSource = (0xff & V) << 8;
                      
                      byte[] source = RAM[kSource >> 13];
                      int sourceP = kSource & 0x1fff;
                      
                      System.arraycopy(source, sourceP, dest, p, 0xA0);
                      // Around 160 ns. XXX - adjust cycles.
                   }
                   return;
      case 0xFF41: V=(byte) ((V&0xF8)|(RAM[LCDSTAT>>13][LCDSTAT&0x1fff]&0x07));
                   break;
      case 0xFF40: ChrGen=(char) (((V&0x10) != 0)? (char) 0x8000:(char) 0x8800);
                   BgdTab=(char) (((V&0x08) != 0)? (char) 0x9C00:(char) 0x9800);
                   WndTab=(char) (((V&0x40) != 0)? (char) 0x9C00:(char) 0x9800);
                   break;
      case 0xFF44: V=0;break;
      case 0xFF47: BPal[0]= (byte) (V&0x03);
                   BPal[1]= (byte) ((V&0x0C)>>2);
                   BPal[2]= (byte) ((V&0x30)>>4);
                   BPal[3]= (byte) ((V&0xC0)>>6);
                   break;
      case 0xFF48: SPal[0][0]= (byte) (V&0x03);
                   SPal[0][1]= (byte) ((V&0x0C)>>2);
                   SPal[0][2]= (byte) ((V&0x30)>>4);
                   SPal[0][3]= (byte) ((V&0xC0)>>6);
                   break;
      case 0xFF49: SPal[1][0]= (byte) (V&0x03);
                   SPal[1][1]= (byte) ((V&0x0C)>>2);
                   SPal[1][2]= (byte) ((V&0x30)>>4);
                   SPal[1][3]= (byte) ((V&0xC0)>>6);
                   break;
      default:     if((A>=0xFF10)&&(A<=0xFF26))
                   {
                     if(SndStream != null)
                     {
                       try {
                         SndStream.write((byte) (A-0xFF10));
                         SndStream.write((byte) V);
                       } catch (IOException e){
                        e.printStackTrace();
                        SndStream = null;
                       }
                     }
                     Sound(A-0xFF10,(byte) V);
                   }
    }
    RAM[A>>13][A&0x1fff]=(byte) V;
  }
  
  void Sound(int reg, byte value){
  }
  
  int SIOSend(){
    return 0;
  }
  
  int SIOReceive(){
    return 0;
  }

  int startGB(URL CartName)
  {
    System.err.println("Opening " + CartName);
    String CartTypes[] =
    {
      "ROM ONLY","ROM+MBC1","ROM+MBC1+RAM",
      "ROM+MBC1+RAM+BATTERY","UNKNOWN",
      "ROM+MBC2","ROM+MBC2+BATTERY"
    };

    int I,J;
    InputStream F;
    // char *P,S[50];
    char A;

    for(I=0;I<8;I++) Page[I]=RAM[I];

    if(Verbose != 0) System.out.println("OK\nOpening " + CartName + "...");
    try {
      F = CartName.openStream();
    }
    catch ( IOException e) {
      e.printStackTrace();
      if(Verbose != 0) System.out.println("FAILED");
      return(0);
    }

    if(Verbose != 0) System.out.println("reading...");
    try {
      readCompletely(F, RAM[0], 0, 0x2000);
      readCompletely(F, RAM[1], 0, 0x2000);
    }
    catch (IOException e){
      e.printStackTrace();
      if(Verbose != 0) System.out.println("FAILED");
      return(0);
    }

    ROMMap[0]=RAM[0];
    ROMMap[1]=RAM[1];
    ROMBanks=2<<RAM[0][0x0148];
    RAMBanks=(RAM[0][0x0149]&0x03)==3? 4:0;
    char Checksum=(char)((RAM[0][0x014E]<<8) | 0xff & (RAM[0][0x014F]));
    MBCType=RAM[0][0x0147]>3;

    if((RAM[0][0x0147]==4)||(RAM[0][0x0147]>6)) {
      System.out.println("\nError loading cartridge: Unknown cartridge type");
      try {
        F.close();
      }
      catch (IOException e) {
      }
      return(0);
    }

    if(Verbose != 0)
    {
      System.out.print("OK\n  Name: ");
      for(int m = 0; m < 16; m++){
        System.out.print((char) RAM[0][0x0134 + m]);
      }
      System.out.println();
      System.out.println("  Type: " + CartTypes[RAM[0][0x0147]]);
      System.out.println("  ROM Size: " + ROMBanks + "x16kB");
      J=(RAM[0][0x0149]&0x03)*2;J=(J != 0)? (1<<(J-1)):0;
      System.out.println("  RAM Size: " + J + "kB");

      J=0xffff & ((RAM[0][0x014B]<<8) + (0xff & RAM[0][0x014A]));
      String companyCode = getCompanyForCode(J);
      System.out.print("  Manufacturer ID: " + hex(J) );
      System.out.println(" [" + (companyCode != null ? companyCode:"?") + "]");

      System.out.println("  Version Number: " + hex((int) RAM[0][0x014C]));
      System.out.println("  Complement Check: " + hex((int) RAM[0][0x014D]));
      System.out.println("  Checksum: " + hex((int) Checksum));
      J= 0xffff & (RAM[0][0x0103]<<8)+ (0xff & RAM[0][0x0102]);
      System.out.println("  Start Address: " + hex(J));
    }

    Checksum+=(0xff & RAM[0][0x014E])+(0xff & RAM[0][0x014F]);
    for(I=0;I<0x2000;I++) Checksum-=(0xff & RAM[0][I]);
    for(I=0;I<0x2000;I++) Checksum-=(0xff & RAM[1][I]);

    if(Verbose != 0) System.out.print("Loading " + ROMBanks + "x16kB ROM banks:\n.");
    for(I=2;I<ROMBanks*2;I++) {
      ROMMap[I]=new byte[0x2000];
      try {
        readCompletely(F, ROMMap[I],0,0x2000);
      } catch (IOException e) {
        if(Verbose != 0) System.err.print("READ FAILED");
        e.printStackTrace();
        break;
      }
      for(J=0;J<0x2000;J++) Checksum-=(0xff & ROMMap[I][J]);
      if(Verbose != 0) System.out.print('.');
    }
    try {
      F.close();
    }
    catch (IOException e) {
    }
    if(I<ROMBanks) return(0);

    if(CheckCRC&&(Checksum&0xFFFF) != 0)
    { System.out.print("\nError loading cartridge: Checksum is wrong");return(0); }

    if(Verbose!= 0) System.out.println("OK");

    if(RAMBanks != 0 &&!MBCType)
    {
      if(Verbose!= 0) System.out.println("Allocating " + RAMBanks + " RAM banks...");
      for(I=0;I<RAMBanks;I++)
        RAMMap[I]=new byte[0x2000];
      if(Verbose!= 0) System.out.println("OK");
    }

    if((RAM[0][0x0147]!=3)&&(RAM[0][0x0147]!=6)) SaveName=null;
    else {
      String CartNameString = CartName.toString();
      int index = CartNameString.lastIndexOf('.');
      if ( index >= 0 ) {
        SaveName = CartNameString.substring(0,index);
      }
      else {
        SaveName = CartNameString;
      }
      SaveName = SaveName + ".sav";

      if(Verbose != 0) System.out.println("Opening save file" + SaveName + "...");
      F = null;
      try {
        F = (new URL(SaveName)).openStream();
        if(Verbose != 0) System.out.println("reading...");
        J=RAM[0][0x0147]==3? 0x2000:0x0200;
        readCompletely(F, RAMMap[0] != null? RAMMap[0]:Page[5], 0, J);
      }
      catch(Exception e) {
        if(Verbose != 0) System.out.print("FAILED");
        e.printStackTrace();
      }
      if ( F != null ) {
        try {
          F.close();
        }
        catch (IOException e) {
        }
      }
    }

    if(CheatCount>0)
    {
      if(Verbose != 0) System.out.print("Patching cheats into the ROM code:");
      /*
      for(J=0;J<CheatCount;J++)
      {
        A=Cheats[J].Address;
        if(Verbose != 0)
          System.out.println("  at %Xh: %Xh -> %Xh\n",A,Cheats[J].Orig,Cheats[J].Value);
        if(A<0x4000)
        { if(ROMMap[0][A]==Cheats[J].Orig) ROMMap[0][A]=Cheats[J].Value; }
        else
          for(I=0,A-=0x4000;I<ROMBanks;I++)
            if(ROMMap[I][A]==Cheats[J].Orig) ROMMap[I][A]=Cheats[J].Value;
      }
      */
    }

    if(SndName != null)
    {
      if(Verbose != 0) System.out.println("Logging soundtrack to %s..." + SndName);
      try {
        SndStream= new FileOutputStream(SndName);
        if(Verbose != 0) System.out.print("OK");
        // SndStream.write("Gameboy Soundtrack File 1.0\032");
      } catch (IOException e){
        e.printStackTrace();
        try {
          SndStream.close();
        }
        catch (IOException e2) {
        }
      }
        
    }

    if(ROMBanks<3) ROMMask=0;
    else { for(I=1;I<ROMBanks;I<<=1);ROMMask=(byte)( I-1);ROMBank=1; }
    if(RAMMap[0] == null) RAMMask=0;
    else { for(I=1;I<RAMBanks;I<<=1);RAMMask=(byte)(I-1);RAMBank=0; }

    if(RAMMap[0] != null) Page[5]=RAMMap[0];
    if(ROMMap[1] != null){
      Page[2]=ROMMap[1];
      Page[3]=ROMMap[2];
      // was Page[2]+0x2000;
    }

    tstatesPerInterrupt = VPeriod/154/11;
    TStep=32768;TCount=0;

    ChrGen=0x8800;BgdTab=WndTab=0x9800;
    primitiveStore(LCDCONT, 0x81);
    primitiveStore(LCDSTAT, 0x00);
    primitiveStore(CURLINE, 0x00);
    primitiveStore(CMPLINE, 0xFF);
    primitiveStore(IFLAGS, 0x00);
    primitiveStore(ISWITCH, 0x00);
    primitiveStore(TIMECNT, 0x01);
    primitiveStore(TIMEMOD, 0x01);
    primitiveStore(TIMEFRQ, 0xF8);
    primitiveStore(SIODATA, 0x00);
    primitiveStore(SIOCONT, 0x7E);
    SIOBuf = (byte) 0x00;
    SIOFlag = (byte) 0x00;
    SprFlag=0;
    JoyState=(byte) 0xFF;

    for(I=0;I<4;I++) SPal[0][I]=SPal[1][I]=BPal[I]=(byte) I;
    primitiveStore(BGRDPAL, 0xE4);
    primitiveStore(SPR0PAL, 0xE4);
    primitiveStore(SPR1PAL, 0xE4);

    if(Verbose != 0) System.out.println("\nRUNNING ROM CODE...");
    reset();
    // Trace(2);
    execute();

    if(Verbose != 0) System.out.println("EXITED at PC = " + hex(PC())); 
    return(1);
  }
  
  static String hex(int v){
    return Integer.toHexString(v) + "h";
  }
  byte[] LCDStates = { 2,2,3,3,3,3,0,0,0,0,0 };
  byte LCount=0;
  int UCount=1;
  byte ACount=0;
  
  int doInterrupt()
  {
    
    byte J,I;

    primitiveStore(DIVREG, primitiveRead(DIVREG) + 1);
    LCount=(byte) ((LCount+1)%11);
    if(primitiveRead(CURLINE)<144){
      primitiveStore(LCDSTAT, (primitiveRead(LCDSTAT)&0xFC) | LCDStates[LCount]);
    }

    switch(LCount)
    {
      case 0:
        /* Proceeding with line refresh */
        break;
      case 2:
        /* Generating VBlank interrupt */
        if(primitiveRead(CURLINE)==144)
          if((primitiveRead(ISWITCH)& VBL_IFLAG) != 0
            && (primitiveRead(LCDCONT)&0x80) != 0)
          {
            primitiveStore(IFLAGS, (IMask= VBL_IFLAG));
            return(0x0040);
          }
        return(0xFFFF);
      case 7:
        /* Generating HBlank interrupt */
        if(((primitiveRead(LCDSTAT)&0x08) != 0)&&(primitiveRead(CURLINE) <144))
          if((primitiveRead(ISWITCH) & LCD_IFLAG) != 0
            &&(primitiveRead(LCDCONT)&0x80) != 0)
          {
            primitiveStore(IFLAGS, (IMask= LCD_IFLAG));
            return(0x0048);
          }
        return(0xFFFF);
      default:
        /* Only LCD state had to be changed */
        return(0xFFFF);
    }

    int curLine = (primitiveRead(CURLINE)+1)%154;
    // System.err.println(curLine);
    primitiveStore(CURLINE, curLine);
    if((!(UCount != 0))&&(curLine<144)) RefreshLine(curLine);
    SprFlag|=primitiveRead(LCDCONT)&0x02;

    /* Checking for line coincidence */
    J= (byte) (LineDelay ? (curLine+1)%154:curLine);
    if(J!=primitiveRead(CMPLINE)){
      primitiveStore(LCDSTAT, primitiveRead(LCDSTAT) & 0xFB);
      J=0x00;
    }
    else {
      primitiveStore(LCDSTAT, primitiveRead(LCDSTAT)| 0x04);
      J=0x40;
    }
    /* If end of frame reached... */
    if(primitiveRead(CURLINE)==144)
    {
      /* Set VBlank state */
      primitiveStore(LCDSTAT, (primitiveRead(LCDSTAT)&0xFC)|0x01);
      J|=0x10;

      /* Write interrupt timestamp into sound log */
      if(SndStream != null ){
        try {
          SndStream.write((byte)0xFF);
        } catch (IOException e) {
          e.printStackTrace();
        }
      }
      /* Refresh screen if needed */
      if(UCount != 0){
        UCount--;
      }
      else
      {
        if((SprFlag != 0) &&(primitiveRead(LCDCONT)&0x80) != 0) RefreshSprites();
        RefreshScreen();SprFlag=0;UCount=UPeriod;
      }

      /* Generate serial IO interrupt */
      if((SIOFlag&0x01) != 0)
      {
        SIOFlag&=0xFE;
        primitiveStore(SIOCONT, primitiveRead(SIOCONT) & 0x7F);
        if((primitiveRead(ISWITCH)&SIO_IFLAG)!= 0){
          primitiveStore(IFLAGS, primitiveRead(IFLAGS) | SIO_IFLAG);
        }
      }

      /* Update joystick */
      JoyState=Joystick();

      /* Autofire emulation */
      ACount=(byte) ((ACount+1)&0x07);
      if(ACount>3)
      {
        if(AutoA) JoyState|=0x10;
        if(AutoB) JoyState|=0x20;
      }

      /* Assign value to JOYPAD */
      I= (byte) (primitiveRead(JOYPAD)|0xCF);
      if(!((I&0x10) != 0)) primitiveStore(JOYPAD, I&(JoyState|0xF0));
      if(!((I&0x20) != 0)) primitiveStore(JOYPAD, I&((JoyState>>4)|0xF0));
    }

    /* Generating LCD controller interrupt */
    if(((J&primitiveRead(LCDSTAT)) != 0)
      && (primitiveRead(ISWITCH)&LCD_IFLAG) != 0
      && (primitiveRead(LCDCONT)&0x80) != 0){
      primitiveStore(IFLAGS, primitiveRead(IFLAGS) | LCD_IFLAG);
    }

    /* Generating timer interrupt */
    if((primitiveRead(TIMEFRQ)&0x04) != 0)
    {
      TCount+=TStep;
      if((TCount&0xFFFF0000) != 0)
      {
        int L;
        L=primitiveRead(TIMECNT)+(TCount>>16);
        TCount&=0x0000FFFF;
        if((L&0xFFFFFF00) != 0)
        {
          primitiveStore(TIMECNT, primitiveRead(TIMEMOD));
          if((primitiveRead(ISWITCH)&TIM_IFLAG) != 0){
            primitiveStore(IFLAGS, primitiveRead(IFLAGS) | TIM_IFLAG);
          }
        }
        else primitiveStore(TIMECNT, L);
      }
    }

    /* Determining interrupt address */
    if((J= (byte) primitiveRead(IFLAGS)) != 0)
    {
      if((J&EXT_IFLAG)!=0) { IMask=EXT_IFLAG;return(0x0060); }
      if((J&SIO_IFLAG)!=0) { IMask=SIO_IFLAG;return(0x0058); }
      if((J&TIM_IFLAG)!=0) { IMask=TIM_IFLAG;return(0x0050); }
      if((J&LCD_IFLAG)!=0) { IMask=LCD_IFLAG;return(0x0048); }
      if((J&VBL_IFLAG)!=0) { IMask=VBL_IFLAG;return(0x0040); }
    }

    /* No interrupt */
    return(0xFFFF);
  } 
  /** USE_XPAL *************************************************/
  /** When defined, it makes Common.h drivers use XPal[] to   **/
  /** obtain color numbers (when there is no 1:1              **/
  /** correspondence, for example in X11).                    **/
  /*************************************************************/
  /* #define USE_XPAL */

  // #ifdef USE_XPAL
  // #define XPAL(N) XPal[N]
  // #else
  final static byte XPAL(int N){ return (byte) N; }
  // #endif

    // #ifdef USE_XPAL
    // #define PAL(N)  Pal[N]
    // #define WPAL(N) WPal[N]
    //   byte Pal[4],WPal[4];
    //   Pal[0]=XPal[BPal[0]];Pal[1]=XPal[BPal[1]];
    //   Pal[2]=XPal[BPal[2]];Pal[3]=XPal[BPal[3]];
    //   WPal[0]=(XPal+8)[BPal[0]];WPal[1]=(XPal+8)[BPal[1]];
    //   WPal[2]=(XPal+8)[BPal[2]];WPal[3]=(XPal+8)[BPal[3]];
    // #else
    final byte PAL(int N) { return  BPal[N]; }
    final byte WPAL(int N) { return  (byte) (BPal[N]+8); }
    // #endif

  void memset(byte[] buf, int offset, int value, int count){
    byte b = (byte) value;
    for(int i = offset, n = offset + count; i < n; i++){
      buf[i] = b;
    }
  }

  /****************************************************************/
  /*** Refresh a line.                                          ***/
  /****************************************************************/
  void RefreshLine(int Y)
  {
    int Offset;
    int P;
    int T; // into RAM (bgd table)
    int R; // relative to XBuf
    int Z; // relative to ZBuf
    int Z0,D0,X1,X2,Shift;
    int D1;
    
    int cLCDCONT = primitiveRead(LCDCONT);
    int cSCROLLX = primitiveRead(SCROLLX);
    int cSCROLLY = primitiveRead(SCROLLY);
    int cWNDPOSX = primitiveRead(WNDPOSX);
    int cWNDPOSY = primitiveRead(WNDPOSY);
 
    D1=WIDTH*Y+(WIDTH-160)/2;
    R=D1;Z=(D1>>3);

    if((cLCDCONT&0x81)!=0x81)
    {
      // Clearing screen buffer
      memset(XBuf, R,XPAL(0),160);
      // Clearing Z-buffer
      memset(ZBuf, Z,0x00,160/8);
      memset(ZBuf, Z+WIDTH*HEIGHT/8,0x00,160/8);
    }
    else
    {
      Offset=Y+cSCROLLY;
      T=BgdTab+((int)(Offset&0xF8)<<2);
      Offset=(Offset&0x07)<<1;

      Shift=cSCROLLX&0x07;
      R-=Shift;Z--;
      X1=cSCROLLX>>3;

      // Determine how much background is shown
      X2=(cWNDPOSX>7)? cWNDPOSX-7:0;
      if(((~cLCDCONT&0x20) != 0 )
        ||(cWNDPOSY>Y)
        ||(X2>159)) X2=160;
      X2=X1+((X2+7)>>3);

      ZBuf[Z + 0]=0x00;

      for(;X1<=X2;X1++)
      {
        D0=primitiveRead(T+(X1&0x1F));
        if((~cLCDCONT&0x10) != 0){
          // Using upper characterset table.
          // Characters are signed.
          D0 = 0xff & (D0 + 0x80);
        }
        P=ChrGen+(D0<<4)+Offset;
        D0=primitiveRead(P);

        // Modifying Z-buffer
        D1=(int)(D0|primitiveRead(P+1))<<Shift;
        ZBuf[Z] |= (D1>>8);
        ZBuf[Z + 1] |=D1&0xFF;
        ZBuf[Z+WIDTH*HEIGHT/8]=0x00;

        // Modifying screen buffer
        D1=(int)primitiveRead(P+1)<<1;
        D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((int)(D0&0xAA)<<7);
        XBuf[R]=PAL((D1&0xC000)>>14);
        XBuf[R+1]=PAL((D1&0x00C0)>>6 );
        XBuf[R+2]=PAL((D1&0x3000)>>12);
        XBuf[R+3]=PAL((D1&0x0030)>>4 );
        XBuf[R+4]=PAL((D1&0x0C00)>>10);
        XBuf[R+5]=PAL((D1&0x000C)>>2 );
        XBuf[R+6]=PAL((D1&0x0300)>>8 );
        XBuf[R+7]=PAL( D1&0x003      );

        // Moving pointers
        R+=8;Z++;
      }
    }

    X1=(cWNDPOSX>7)? (cWNDPOSX-7):0;
    if(((cLCDCONT&0xA0)==0xA0)&&(cWNDPOSY<=Y)&&(X1<160))
    {
      D1=WIDTH*Y+(WIDTH-160)/2+X1;
      R=D1;Z=(D1>>3);

      Offset=Y-cWNDPOSY;
      T=WndTab+((int)(Offset&0xF8)<<2);
      Offset=(Offset&0x07)<<1;

      D0=0xFF>>(X1&0x07);
      if((cLCDCONT&0x02) != 0) ZBuf[Z+WIDTH*HEIGHT/8]= (byte) (Z0=0x00);
      else { ZBuf[Z+WIDTH*HEIGHT/8]=(byte) D0;Z0=0xFF; }
      ZBuf[Z]|=D0;Z++;

      for(X1>>=3;X1<(160>>3);X1++)
      {
        // Window characters are signed.
        
        D0= 0xff & ( primitiveRead(T++)+0x80 );
        P=0x8800+((int)D0<<4)+Offset;

        // Modifying Z-buffer
        ZBuf[Z+WIDTH*HEIGHT/8]= (byte) Z0; ZBuf[Z] = (byte) 0xFF;

        // Modifying screen buffer
        D0=primitiveRead(P);D1=(int)primitiveRead(P+1)<<1;
        D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((int)(D0&0xAA)<<7);
        XBuf[R+0]=WPAL((D1&0xC000)>>14);
        XBuf[R+1]=WPAL((D1&0x00C0)>>6 );
        XBuf[R+2]=WPAL((D1&0x3000)>>12);
        XBuf[R+3]=WPAL((D1&0x0030)>>4 );
        XBuf[R+4]=WPAL((D1&0x0C00)>>10);
        XBuf[R+5]=WPAL((D1&0x000C)>>2 );
        XBuf[R+6]=WPAL((D1&0x0300)>>8 );
        XBuf[R+7]=WPAL( D1&0x0003     );

        // Moving pointers
        R+=8;Z++;
      }
    }
  }


    /* Conversion matrix where result has bits */
    /* flipped over and inverted.              */
    int[] Conv =
    {
      0xFF,0x7F,0xBF,0x3F,0xDF,0x5F,0x9F,0x1F,0xEF,0x6F,0xAF,0x2F,
      0xCF,0x4F,0x8F,0x0F,0xF7,0x77,0xB7,0x37,0xD7,0x57,0x97,0x17,
      0xE7,0x67,0xA7,0x27,0xC7,0x47,0x87,0x07,0xFB,0x7B,0xBB,0x3B,
      0xDB,0x5B,0x9B,0x1B,0xEB,0x6B,0xAB,0x2B,0xCB,0x4B,0x8B,0x0B,
      0xF3,0x73,0xB3,0x33,0xD3,0x53,0x93,0x13,0xE3,0x63,0xA3,0x23,
      0xC3,0x43,0x83,0x03,0xFD,0x7D,0xBD,0x3D,0xDD,0x5D,0x9D,0x1D,
      0xED,0x6D,0xAD,0x2D,0xCD,0x4D,0x8D,0x0D,0xF5,0x75,0xB5,0x35,
      0xD5,0x55,0x95,0x15,0xE5,0x65,0xA5,0x25,0xC5,0x45,0x85,0x05,
      0xF9,0x79,0xB9,0x39,0xD9,0x59,0x99,0x19,0xE9,0x69,0xA9,0x29,
      0xC9,0x49,0x89,0x09,0xF1,0x71,0xB1,0x31,0xD1,0x51,0x91,0x11,
      0xE1,0x61,0xA1,0x21,0xC1,0x41,0x81,0x01,0xFE,0x7E,0xBE,0x3E,
      0xDE,0x5E,0x9E,0x1E,0xEE,0x6E,0xAE,0x2E,0xCE,0x4E,0x8E,0x0E,
      0xF6,0x76,0xB6,0x36,0xD6,0x56,0x96,0x16,0xE6,0x66,0xA6,0x26,
      0xC6,0x46,0x86,0x06,0xFA,0x7A,0xBA,0x3A,0xDA,0x5A,0x9A,0x1A,
      0xEA,0x6A,0xAA,0x2A,0xCA,0x4A,0x8A,0x0A,0xF2,0x72,0xB2,0x32,
      0xD2,0x52,0x92,0x12,0xE2,0x62,0xA2,0x22,0xC2,0x42,0x82,0x02,
      0xFC,0x7C,0xBC,0x3C,0xDC,0x5C,0x9C,0x1C,0xEC,0x6C,0xAC,0x2C,
      0xCC,0x4C,0x8C,0x0C,0xF4,0x74,0xB4,0x34,0xD4,0x54,0x94,0x14,
      0xE4,0x64,0xA4,0x24,0xC4,0x44,0x84,0x04,0xF8,0x78,0xB8,0x38,
      0xD8,0x58,0x98,0x18,0xE8,0x68,0xA8,0x28,0xC8,0x48,0x88,0x08,
      0xF0,0x70,0xB0,0x30,0xD0,0x50,0x90,0x10,0xE0,0x60,0xA0,0x20,
      0xC0,0x40,0x80,0x00
    };
    
  /****************************************************************/
  /*** Refresh sprites.                                         ***/
  /****************************************************************/

  // #undef PAL
  // #ifdef USE_XPAL
  // #define PAL(N) Pal[N]
  //   byte Pal[4];
  // #else
  // #define PAL(N) R[N]+4
  // #endif
  
  final byte PAL(int R, int N){
    return (byte) (SPal[R][N]+4);
  }

  void RefreshSprites()
  {
    int P; // relative to XBuf
    int T; // relative to RAM
    int S; // relative to RAM
    int R; // relative to SPal
    int Z; // relative to ZBuf
    int I;
    int J;
    int K;
    int N;
    int D0;
    int D1; // unsigned?
    int Shift,DY;


    N= (primitiveRead(LCDCONT)&0x04 ) != 0? 16:8;

    for(S=0xFE9C,J=40;J != 0;S-=4,J--)
      if( primitiveRead(S) != 0
        && (primitiveRead(S)<160)
        && primitiveRead(S+1) != 0
        &&(primitiveRead(S+1)<168))
      {
        if(primitiveRead(S)<16) { D0=0;K=16-primitiveRead(S);I=(K<N)? N-K:0; }
        else {
          D0=primitiveRead(S)-16;
          K=0;
          I=(160-primitiveRead(S)<N)? 160-primitiveRead(S):N;
        }
        Shift=primitiveRead(S+1)-(((primitiveRead(S+3)&0x20 ) != 0 )? 1:8);
        D1=D0*WIDTH+(WIDTH-160)/2+Shift;
        P=D1;Z=(D1>>3);
        Shift&=0x07;
        DY=2;
        if((primitiveRead(S+3)&0x40) != 0) { DY=-DY;K=N-K-1; }
        R= ((primitiveRead(S+3)&0x10) != 0)? 1:0;
  // #ifdef USE_XPAL
  //       Pal[0]=(XPal+4)[R[0]];Pal[1]=(XPal+4)[R[1]];
  //       Pal[2]=(XPal+4)[R[2]];Pal[3]=(XPal+4)[R[3]];
  // #endif
        T=0x8000+(((N>8)? primitiveRead(S+2)&0xFE:primitiveRead(S+2))<<4)+(K<<1);

        // If on top of background, clip only against the window
        if((~primitiveRead(S+3)&0x80) != 0) Z+=WIDTH*HEIGHT/8;

        if((primitiveRead(S+3)&0x20) != 0)
        {
          Shift=7-Shift;
          for(;I != 0;I--,P+=WIDTH,Z+=WIDTH/8,T+=DY)
          {
            D0=primitiveRead(T);D1=primitiveRead(T+1);
            K=(((int)ZBuf[Z-1]<<8)|ZBuf[Z])>>Shift;
            K=(D0|D1)&Conv[K];
            if(K != 0)
            {
              D1<<=1;
              D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((int)(D0&0xAA)<<7);
              if((K&0x80)!=0) XBuf[P]=PAL(R, (D1&0xC000)>>14);
              if((K&0x40)!=0) XBuf[P-1]=PAL(R, (D1&0x00C0)>>6 );
              if((K&0x20)!=0) XBuf[P-2]=PAL(R, (D1&0x3000)>>12);
              if((K&0x10)!=0) XBuf[P-3]=PAL(R, (D1&0x0030)>>4 );
              if((K&0x08)!=0) XBuf[P-4]=PAL(R, (D1&0x0C00)>>10);
              if((K&0x04)!=0) XBuf[P-5]=PAL(R, (D1&0x000C)>>2 );
              if((K&0x02)!=0) XBuf[P-6]=PAL(R, (D1&0x0300)>>8 );
              if((K&0x01)!=0) XBuf[P-7]=PAL(R, D1&0x0003     );
            }
          }
        }
        else
        {
          Shift=8-Shift;
          for(;I != 0;I--,P+=WIDTH,Z+=WIDTH/8,T+=DY)
          {
            D0=primitiveRead(T);D1=primitiveRead(T+1);
            K=(((int)ZBuf[Z]<<8)|ZBuf[Z+1])>>Shift;
            K=(D0|D1)&~K;
            if(K != 0)
            {
              D1<<=1;
              D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((int)(D0&0xAA)<<7);
              if((K&0x80) != 0 ) XBuf[P]=PAL(R, (D1&0xC000)>>14);
              if((K&0x40) != 0 ) XBuf[P+1]=PAL(R, (D1&0x00C0)>>6 );
              if((K&0x20) != 0 ) XBuf[P+2]=PAL(R, (D1&0x3000)>>12);
              if((K&0x10) != 0 ) XBuf[P+3]=PAL(R, (D1&0x0030)>>4 );
              if((K&0x08) != 0 ) XBuf[P+4]=PAL(R, (D1&0x0C00)>>10);
              if((K&0x04) != 0 ) XBuf[P+5]=PAL(R, (D1&0x000C)>>2 );
              if((K&0x02) != 0 ) XBuf[P+6]=PAL(R, (D1&0x0300)>>8 );
              if((K&0x01) != 0 ) XBuf[P+7]=PAL(R,  D1&0x0003     );
            }
          }
        }
      }
  }


  /****************************************************************/
  /*** Display the screen buffer.                               ***/
  /****************************************************************/
  void RefreshScreen() { 
    // System.err.println("Refreshing screen");
    /* PutImage(); */
    if ( fMyObserver != null ) {
      fMyObserver.refreshScreen(this);
    }
    delay();
  }

  void setObserver(GameboyObserver o){
    fMyObserver = o;
  }
  Date last;
  void delay(){
    Date now = new Date();
    if ( last != null ) {
      int ms = (int) (now.getTime() - last.getTime());
      int targetMS = 1000 / 30;
      // Need to sleep at least 5 ms, or else the
      // debugger locks up on Visual Cafe under NT 4.0
      
      int delay = Math.max(10, targetMS - ms);
      try {
        Thread.currentThread().sleep(delay);
      }
      catch (Throwable t)
      {
      }
      // System.err.println(ms + " ms (Target " + targetMS + " ms )");
    }
    last = now;
  }

  public final static int JOY_START = 128;
  public final static int JOY_SELECT = 64;
  public final static int JOY_B = 32;
  public final static int JOY_A = 16;
  public final static int JOY_D = 8;
  public final static int JOY_U = 4;
  public final static int JOY_L = 2;
  public final static int JOY_R = 1;
  
  private byte HumanJoyState = (byte) 0xff;
  
  public void setJoystick(int state){
    HumanJoyState = (byte) state;
  }
  private byte Joystick() {
    return HumanJoyState;
  }
 
  final void primitiveStore(char addr, int val){
    RAM[addr>>13][addr&0x1fff] = (byte) val;
  }
  
  final int primitiveRead(int addr){
    return 0xff & RAM[addr>>13][addr&0x1fff];
  }
  
  void readCompletely(InputStream in, byte[] buf, int index, int length)
    throws IOException
  {
    while(length > 0 ){
      int read = in.read(buf, index, length);
      if ( read < 0 ) {
        throw new IOException("EOF");
      }
      length -= read;
      index += read;
    }
  }
  
  static String hex2(int v){
    return Debug.hex2(v);
  }
  
  static String hex4(int v){
    return Debug.hex4(v);
  }  
}